#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include <math.h>


#define  ACCUM_SIZE  16       // Passed in as -D gcc parameter. 16 64 (Min=8+WORD_POSITION, Max=64) Bit width of generated random numbers.
//#define  PARITY_BIT  0        // Bit number of summed result to substitute with summing parity calculation.

#define  CONSTANT_A  14       // Passed in as -D gcc parameter. 14 55 24 24
#define  CONSTANT_B  2        // Passed in as -D gcc parameter.  2 14 16 16
#define  CONSTANT_C  7        // Passed in as -D gcc parameter.  7 36 37 37
#define  CONSTANT_D  15       // Passed in as -D gcc parameter.  5    29  5
//#define  CONSTANT_E  7        // Passed in as -D gcc parameter.           7
//#define  CONSTANT_F  9        // Passed in as -D gcc parameter.           9


#if ACCUM_SIZE > 32
	#define  ACCUM_MASK  (((1ULL<<(ACCUM_SIZE-1))-1)<<1|1)  // Bit of a shuffle to work on all 64 bits :)
	#define  SAMPLE_MASK  (((1ULL<<(SAMPLE_SIZE-1))-1)<<1|1)  // ditto
	typedef  uint64_t  rngword_t;
#elif ACCUM_SIZE > 16
	#define  ACCUM_MASK  (((1UL<<(ACCUM_SIZE-1))-1)<<1|1)   // max 32 bits
	#define  SAMPLE_MASK  (((1UL<<(SAMPLE_SIZE-1))-1)<<1|1)
	typedef  uint32_t  rngword_t;
#else
	#define  ACCUM_MASK  (((1U<<(ACCUM_SIZE-1))-1)<<1|1)    // max 16 bits
	#define  SAMPLE_MASK  (((1U<<(SAMPLE_SIZE-1))-1)<<1|1)
	typedef  uint16_t  rngword_t;
#endif


//#include "xoroshiro128pcg.c"
#include "xoroshiro128lcg.c"
//#include "xoroshiro128parity.c"



int  main(int argc, char* argv[])
{
	uint32_t  freqs[256], irng, ipair, next_prn, prev_prn;
	uint8_t  *pairs, ifreq;
	FILE  *fh;
	struct timespec  start, stop;
	double  duration;



	// initialize pair array
	pairs = (uint8_t *)calloc( 1UL<<32, sizeof(uint8_t) );    // Zero'd 4 GB memory allocation
	if( pairs == NULL )
	{
		printf( "allocation error\n" );
		exit(1);
	}

	// initialize prev_prn
	prev_prn = next();

	clock_gettime( CLOCK_MONOTONIC, &start );
	printf( "Start at %.3f\n", start.tv_sec + start.tv_nsec/1.0e9 );


	// iterate full period and increment pair values
	for( irng = 0; irng < 1UL<<32-1; irng++ )
	{
		next_prn = next();
		ipair = (next_prn << 16) | prev_prn;   // Little-endian
		pairs[ipair]++;
		prev_prn = next_prn;
	}

	clock_gettime( CLOCK_MONOTONIC, &stop );
	duration = (time_t)(stop.tv_sec - start.tv_sec) + (long)(stop.tv_nsec - start.tv_nsec)/1.0e9;
	printf( "Paired at %.3f  -  Duration = %.3f\n", stop.tv_sec + stop.tv_nsec/1.0e9, duration );


	// initialize freq array
	for( irng = 0; irng < 256; irng++ )
		freqs[irng] = 0;

	// read pair array and increment freq values
	for( irng = 0; irng < 1UL<<32-1; irng++ )
	{
		ifreq = pairs[irng];
		freqs[ifreq]++;
	}

	clock_gettime( CLOCK_MONOTONIC, &stop );
	duration = (time_t)(stop.tv_sec - start.tv_sec) + (long)(stop.tv_nsec - start.tv_nsec)/1.0e9;
	printf( "Finish at %.3f  -  Total duration = %.3f\n", stop.tv_sec + stop.tv_nsec/1.0e9, duration );


	// write freq array to file
	fh = fopen( "dist_report.bin", "a" );
	if( fh != NULL )
	{
		fwrite( freqs, sizeof(freqs), 1, fh );
		fclose( fh );
	}
	else
		printf( "bin file error\n" );

	fh = fopen( "dist_report.txt", "a" );
	if( fh != NULL )
	{
		fprintf( fh, "Frequency distribution report for Xoroshiro32++ [%2d %2d %2d %2d]\n", CONSTANT_A, CONSTANT_B, CONSTANT_C, CONSTANT_D );

		for( irng = 0; irng < 256; irng++ )
			fprintf( fh, "%3d  %d\n", irng, freqs[irng] );

		fprintf( fh, "\n" );
		fclose( fh );
	}
	else
		printf( "txt file error\n" );


	free( pairs );
	return 0;
}

